### AddFields.py ###
# Adds fields to 
# Excel Columns Expected to be in the following order:
#  SDE Path	                                              Field Name	  Field Type	Field Precision (Optional)	Field Scale (Optional)	Field Length (Optional)	Field Alias (Optional)	Not Nullable (Optional)	Field Domain (Optional)
#  <FC or Table Name>       <fieldName>       TEXT                                                            20
#   
# SDE path must be full path to .sde connection file.
# Excel doc must be full path.
# Usage: python.exe <Path>\AddFields.py "<Path>\MyExcelDoc.xls" "<Path>\MyGDBConnect.sde"

import arcpy
import sys
from xlrd import open_workbook
import logging
import time
import os

# Add Field function
def addField (inTable,fieldName,fieldType,fieldPrecision,fieldScale,fieldLength,fieldAlias,fieldIsNullable,fieldDomain):  
  # Check if fc/table exists
  if not arcpy.Exists(inTable):
    msg = 'FeatureClass/Table ' + inTable + ' does not exist!'
    logging.info(msg)
    print msg
    return    
  # Check if field exists in fc/table 
  field_names = [f.name for f in arcpy.ListFields(inTable)]
  #if fieldName in field_names:
  if fieldName.lower() in [x.lower() for x in field_names]:
    msg = 'Field ' + fieldName + ' already exists in FeatureClass/Table ' + inTable
    logging.info(msg)
    print msg
    return

  # Try to Add Field
  try:
    # Setup parameters
    fieldIsNull = None
    if fieldIsNullable:
      if fieldIsNullable.lower() == 'no':
        fieldIsNull = 'NON_NULLABLE'
    fieldPrec = None
    if fieldPrecision:
      fieldPrec = long(fieldPrecision)
    fieldScl = None
    if fieldScale:
      fieldScl = long(fieldScale)
    fieldLen = None
    if fieldLength:
      fieldLen = long(fieldLength)
    arcpy.AddField_management(inTable, fieldName, fieldType, fieldPrec, fieldScl, fieldLen, fieldAlias, fieldIsNull,field_domain=fieldDomain)
    msg = "Added Field: " + field_name + ' to ' + inTable
    logging.info(msg)
    print msg
  except Exception as err:
    msg = 'Error running AddField_management! ' + err.args[0] + '  FeatureClass/Table:' + inTable + '  Field:' + fieldName
    logging.error(msg)
    print msg
                  

def processDataset(inTable,fieldName,fieldType,fieldPrecision,fieldScale,fieldLength,fieldAlias,fieldIsNullable,fieldDomain):
  msg = "PROCESSING DATASET: " + inTable
  logging.info(msg)
  print msg  
  skipNames = []
  # Look for any geometricNetworks to get name
  dataSetList = arcpy.ListDatasets()
  if dataSetList:
    for ds in arcpy.ListDatasets():
      desc = arcpy.Describe(ds)
      if hasattr(desc, "name"):
        print("SubDataset Name:        " + desc.name)
      if hasattr(desc, "dataType"):
        print("DataType:    " + desc.dataType)
        if desc.dataType == 'GeometricNetwork':
          skipNames.append(desc.name)
  # Loop through all FC items in dataset
  fcList = arcpy.ListFeatureClasses()
  if fcList:
    msg = 'Processing field: ' + fieldName
    logging.info(msg)
    print msg
    for fc in arcpy.ListFeatureClasses():     
      fcPath = os.path.join(inTable, fc)
      #print(fcPath)    
      # Check if fc name begins with GeometricNetwork name and skip if needed.
      for skipName in skipNames:
        if fc.startswith(skipName):
          msg = 'Skipping Feature Class ' + fc + '. Part of '+ skipName + ' GeometricNetowrk.'
          logging.info(msg)
          print msg 
          continue
      # Call addField       
      addField (fcPath,fieldName,fieldType,fieldPrecision,fieldScale,fieldLength,fieldAlias,fieldIsNullable,fieldDomain)
    

    
excel_doc = arcpy.GetParameterAsText(0)
sde_conn = arcpy.GetParameterAsText(1)

try:
    # Set the parameters here in the script if not passing in through command
    # line arguments.
    #sde_conn = r"C:\Users\Myprofile\AppData\Roaming\ESRI\Desktop10.5\ArcCatalog\myconn.ag4i.sde"
    #excel_doc = r"C:\Users\Myprofile\Documents\EMCS\AddFields.xlsx"
    if not excel_doc:
            excel_doc = raw_input("Please provide path to Excel doc: ")

    if not sde_conn:
        sde_conn = raw_input("Please provide path to SDE connection file: ")
    
    # Check excel_doc and try parse out theme name
    if excel_doc:
        excel_filename = os.path.basename(excel_doc)
        # Get string before the underscore
        theme_name = excel_filename.split('_')[0]
    # Setup Logging
    timestr = time.strftime("%Y%m%d-%H%M%S")
    logFileName = 'AddFields_' + theme_name + '_' + timestr + '.log'
    logging.basicConfig(filename=logFileName, level=logging.INFO, format='%(asctime)s %(message)s')
    logging.info('Started')
    print 'Started'


    if not os.path.isabs(excel_doc):
        wkdir = os.path.dirname(os.path.realpath(__file__))
        excel_doc = os.path.join(wkdir, excel_doc)
        
    #Verify workspace exists
    if not arcpy.Exists(sde_conn):
        msg = "Error: Workspace {0} does not exist!".format(sde_conn)
        logging.error(msg)
        print msg
        sys.exit()

    #Verify Excel doc exists
    if not os.path.exists(excel_doc):
        msg = "Error: Excel document {} does not exist!".format(excel_doc)
        logging.error(msg)
        print msg
        sys.exit() 
        
    logging.info('Excel doc: ' + excel_doc)
    print 'Excel doc: ' + excel_doc

    # Read Excel Sheet
    wb = open_workbook(excel_doc)
    for sheet in wb.sheets():
        # Only use sheet 'AddFields'
        if sheet.name != 'AddFields':
          continue
        number_of_rows = sheet.nrows
        number_of_columns = sheet.ncols
        logging.info(' ')
        print ' '
        logging.info('Processing Excel sheet: ' + sheet.name)
        print 'Processing Excel sheet: ' + sheet.name
        if number_of_rows > 1:
          print "Rows: " + str(number_of_rows)
          
          # Read column values from row
          for row in range(1, number_of_rows):
              in_table = (sheet.cell(row,0).value)
              field_name = (sheet.cell(row,1).value)
              field_type = (sheet.cell(row,2).value)
              field_precision = (sheet.cell(row,3).value)
              field_scale = (sheet.cell(row,4).value)
              field_length = (sheet.cell(row,5).value)
              field_alias = (sheet.cell(row,6).value)
              field_is_nullable = (sheet.cell(row,7).value)
              field_domain = (sheet.cell(row,8).value)              
              
              # Put together full sde path.
              in_table = os.path.join(sde_conn, in_table)

              # Check if db object exists
              if not arcpy.Exists(in_table):
                msg = "Skipping " + in_table + "! Data object does not exist."
                logging.info(msg)
                print msg  
                continue

              # Check db object type from path
              desc = arcpy.Describe(in_table)
              if hasattr(desc, "dataType"):
                print("DataType:    " + desc.dataType)
              else:
                msg = "Skipping " + in_table + "! Data object does not have a dataType attribute."
                logging.info(msg)
                print msg  
              
              arcpy.env.workspace = in_table
              logging.info('Workspace: ' + in_table)

              # Check type of workspace and handle appropriately
              fcTable = ['FeatureClass', 'Table']
              if desc.dataType in fcTable:
                msg = 'PROCESSING FC/TABLE: ' + in_table
                logging.info(msg)
                print msg
                msg = 'Processing field: ' + field_name
                logging.info(msg)
                print msg 
                addField(in_table, field_name, field_type, field_precision, field_scale, field_length, field_alias, field_is_nullable, field_domain)
              elif desc.dataType == 'FeatureDataset':
                  processDataset(in_table, field_name, field_type, field_precision, field_scale, field_length, field_alias, field_is_nullable, field_domain)  
              elif desc.dataType == 'Workspace':
                msg = 'PROCESSING WORKSPACE: ' + in_table
                logging.info(msg)
                print msg 
                ws = in_table
                datasets = arcpy.ListDatasets(feature_type='feature')
                for ds in datasets:
                  dsPath = os.path.join(ws, ds)
                  arcpy.env.workspace = dsPath
                  processDataset(dsPath, field_name, field_type, field_precision, field_scale, field_length, field_alias, field_is_nullable, field_domain)
                
                # Set workspace back to root workspace
                arcpy.env.workspace = in_table
                msg = 'Processing field: ' + field_name
                logging.info(msg)
                print msg
                for fc in arcpy.ListFeatureClasses():
                  fcPath = os.path.join(arcpy.env.workspace, fc)
                  addField(fcPath, field_name, field_type, field_precision, field_scale, field_length, field_alias, field_is_nullable, field_domain)
                for tbl in arcpy.ListTables():
                  tblPath = os.path.join(arcpy.env.workspace, tbl)
                  addField(tblPath, field_name, field_type, field_precision, field_scale, field_length, field_alias, field_is_nullable, field_domain)
              else:
                logging.info('Not a recognized datatype')
                print 'Not a recognized datatype'
                continue             
        else:
          logging.warning('No rows found in excel doc!')

    logging.info('Finished')
    print 'Finished'
 
except Exception as err:
    import traceback, sys
    tb = sys.exc_info()[2]
    logging.error("Line %i" % tb.tb_lineno)
    print "Line %i" % tb.tb_lineno
    logging.error(err.message)
    print err.message
    
