### DomainOperations.py ###
# Assigns Domains to Fields using Excel data
# Excel Columns Expected to be in the following order:
#  Feature Class Name 	 Field Name  	  Domain Name 	      Subtypes     Operation      New Domain Name   Code    Code Description    Coded Value To Remove       Update Records
#   myFeatureClass1	     myField1	      myDomainName1	                  Update Code                        0        MyDescription             1                       Yes
#   myFeatureClass2	     myField2  	      myDomainName2        All          Assign        myNewDomainName                                                               Yes
# Workspace must be full path to .sde connection file
# Excel doc must be full path or just name of file if it is in the same directory as the script.
# (Subtypes can be empty, but column needs to be present)
# (If New Domain Name is provided then the new domain name is added.
# Usage: python.exe <Path>\DomainOperations.py "<Path>\MyExcelDoc.xls" "<Path>\MySDEConnFile.sde"
# You must use Schema Owner connection file

import arcpy
import sys, os
from xlrd import open_workbook
import logging
import time
import errno


def LogMsg(msg, bPrintMsg = True):
  logging.info(msg)
  if bPrintMsg: 
    print '{0}'.format(msg) + '\r\n'


def AssignDomain(in_table, field_name, domain_name, subtype_code):
    # Check if in_table exists if not continue to next row
    if not arcpy.Exists(in_table):
        msg = 'FeatureClass/Table ' + in_table + ' does not exist!'
        LogMsg(msg)
        return

    print "Working on FeatureClass:" + in_table + '  Field:' + field_name + '  DomainName:' + domain_name
    if (subtype_code.lower() == 'none') or (subtype_code == '') or (not subtype_code):
        msg = 'AssigningDomainToField   FeatureClass:' + in_table + '  Field:' + field_name + '  DomainName:' + domain_name
        LogMsg(msg)
    try:
        arcpy.AssignDomainToField_management(in_table, field_name, domain_name)
    except Exception as err:
        msg = 'Error running AssignDomainToField_management! {0}  FeatureClass:{1}  Field:{2}  DomainName:{3}'.format(err.args[0], in_table, field_name, domain_name)
        logging.error(msg)
        print msg
        
    else:
        # if All, Get list of subtypes for FeatureClass and create
        # subtype list
        if subtype_code.lower() == 'all':                  
            subtypes = arcpy.da.ListSubtypes(in_table)                  
            if not subtypes:
                print "Subtypes is null"                  
            # Verify we have subtypes else just assign domain
            if len(subtypes.values()[0]['SubtypeField']) > 0:
                for stcode, stdict in list(subtypes.items()):
                    msg = 'AssigningDomainToField   FeatureClass:{0}  Field:{1}  DomainName:{2}  Subtype code:{3}'.format(in_table,field_name,domain_name,str(stcode))
                    LogMsg(msg)
                    try:
                        arcpy.AssignDomainToField_management(in_table, field_name, domain_name, stcode)
                    except Exception as err:
                        logging.error('Error running AssignDomainToField_management! ' + err.args[0] + '  FeatureClass:' + in_table + '  Field:' + field_name + '  DomainName:' + domain_name + '  Subtype code:' + str(stcode))
            else:
                msg = 'AssigningDomainToField   FeatureClass:{0}  Field:{1}  DomainName:{2}'.format(in_table,field_name,domain_name)
                LogMsg(msg)
                try:
                    arcpy.AssignDomainToField_management(in_table, field_name, domain_name)
                except Exception as err:
                    logging.error('Error running AssignDomainToField_management! ' + err.args[0] + ' FeatureClass:' + in_table + '  Field:' + field_name + '  DomainName:' + domain_name)
                    print 'Error running AssignDomainToField_management! ' + err.args[0] + ' FeatureClass:' + in_table + '  Field:' + field_name + '  DomainName:' + domain_name


def AddDomain(in_workspace, domain_name, domain_description, field_type):
    # Verify domain does not exist before attempting to add
    domain_exists = False
    domains = arcpy.da.ListDomains(in_workspace)
    for domain in domains:
        if domain.domainType == 'CodedValue' and domain.name == domain_name:
            domain_exists = True
            break
    if not domain_exists:
        # Add coded value
        result = arcpy.CreateDomain_management(in_workspace, domain_name, domain_description, field_type)
        msg = "CreateDomain result: {}".format(result)
        LogMsg(msg)
    else:
        msg = "Domain {} already exists.".format(domain_name)
        LogMsg(msg)

def AddCodedValue(in_workspace, domain_name, coded_value_add, coded_desc_add):
    LogMsg("Adding coded value {0} to {1} domain...".format(coded_value_add, domain_name))     
    # Add coded value
    result = arcpy.AddCodedValueToDomain_management(in_workspace, domain_name, coded_value_add, coded_desc_add)
    msg = "AddCodedValueToDomain result: {}".format(result)
    LogMsg(msg)

def UpdateDomainCode(in_workspace, in_table, field_name, domain_name, coded_value_add, coded_desc_add, coded_value_remove, update_records):
    LogMsg("Updating domain...")    
     
    # Add coded value
    result = arcpy.AddCodedValueToDomain_management(in_workspace, domain_name, coded_value_add, coded_desc_add)
    msg = "AddCodedValueToDomain result: {}".format(result)
    LogMsg(msg)    
    
    # Update records if needed
    if update_records == "Yes" and coded_value_remove:
        # Put together full sde path.
        in_table = os.path.join(in_workspace, in_table)

        # Check if db object exists
        if not arcpy.Exists(in_table):
            msg = "Skipping updating domain records for FC/Table {0}! Data object does not exist.".format(in_table)
            LogMsg(msg)
            return      

        # Check if field exists
        field_exists = False
        fields = arcpy.ListFields(in_table)
        for field in fields:
            if field.name.lower() == field_name.lower():
                field_exists = True
                break
        
        if not field_exists:
            msg = "Skipping domain records update. Field {0} does not exist in {1}".format(field_name,in_table)
            logging.warn(msg)
            print msg
            return
       
        #Iterate through the records and update domain value if old coded value is found
        msg = "Updating records with new code value..."
        LogMsg(msg)
        where_clause = field_name + " = '" + coded_value_remove + "'"
        with arcpy.da.UpdateCursor(in_table,field_name,where_clause) as cursor:
            for row in cursor:
                if row[0] == coded_value_remove:
                    row[0] = coded_value_add
                # Update the cursor with the updated list
                cursor.updateRow(row)

    # Remove coded value
    if coded_value_remove:
        # Process: Remove a coded value from the domain
        msg = "Romoving coded value: {}".format(coded_value_remove)
        LogMsg(msg)
        # Verify coded_value_remove exists before attempting to remove
        coded_value_remove_exists = False
        domains = arcpy.da.ListDomains(in_workspace)
        for domain in domains:
            if domain.domainType == 'CodedValue' and domain.name == domain_name:
                coded_values = domain.codedValues
                for val, desc in coded_values.items():
                    print('{0} : {1}'.format(val, desc))
                    if val == coded_value_remove:
                        msg = "Coded value to remove {} exists!".format(val)
                        LogMsg(msg)
                        coded_value_remove_exists = True

        # Remove coded value if it exists
        try:
            if coded_value_remove_exists:
                result = arcpy.DeleteCodedValueFromDomain_management(in_workspace, domain_name, coded_value_remove)
            else:
                msg = "Coded value to remove {} does not exist!".format(coded_value_remove)
                LogMsg(msg)
        except Exception as err:
            msg = "Error: Could not DeleteCodedValueFromDomain! Domain Name: {0}  Coded Value to Remove: {1}  Exception: {2}".format(domain_name,coded_value_remove,err.args[0])
            LogMsg(msg)
            return

        msg = "DeleteCodedValueFromDomain result: {}".format(result)
        LogMsg(msg)


def RemoveDomain():
    print "Removing domain...not yet implemented!"
    # Process: Remove the constraint from the material field
    #arcpy.RemoveDomainFromField_management(inFeatures, inField) 
    # Process: Delete the domain
    #arcpy.DeleteDomain_management(dWorkspace, domName)


def main(argv):
    excel_doc = None
    wrkspc = None

    if len(argv) > 1:
      excel_doc = argv[1]
    if len(argv) > 2:
      wrkspc = argv[2]


    try:
        # Set the parameters here in the script if not passing in through command
        # line arguments.
        #wrkspc = r"C:\Users\Myprofile\AppData\Roaming\ESRI\Desktop10.5\ArcCatalog\my.sde"
        #excelDoc = r"C:\Users\Myprofile\Documents\EMCS\DomainsChange.xlsx"

        if not excel_doc:
            excel_doc = raw_input("Please provide path to Excel doc: ")

        if not wrkspc:
            wrkspc = raw_input("Please provide path to SDE connection file: ")
        # Setup Logging
        # Check excel_doc and try parse out theme name
        if excel_doc:
            excel_filename = os.path.basename(excel_doc)
            # Get string before the underscore
            theme_name = excel_filename.split('_')[0]
        timestr = time.strftime("%Y%m%d-%H%M%S")
        logFileName = 'DomainOperations_' + theme_name + '_' + timestr + '.log'
        logging.basicConfig(filename=logFileName, level=logging.INFO, format='%(asctime)s %(message)s')
        logging.info('Started')
        print 'Started'

        if not os.path.isabs(excel_doc):
            wkdir = os.path.dirname(os.path.realpath(__file__))
            excel_doc = os.path.join(wkdir, excel_doc)
        
        #Verify workspace exists
        if not arcpy.Exists(wrkspc):
            msg = "Error: Workspace {0} does not exist!".format(wrkspc)
            logging.error(msg)
            print msg
            return

        #Verify Excel doc exists
        if not os.path.exists(excel_doc):
            msg = "Error: Excel document {} does not exist!".format(excel_doc)
            logging.error(msg)
            print msg
            return

        logging.info('Workspace: ' + wrkspc)
        print 'Workspace: ' + wrkspc
        logging.info('Excel doc: ' + excel_doc)
        print 'Excel doc: ' + excel_doc

        ## Verify workspace exists
        #if not arcpy.Exists(wrkspc):
        #    msg = "Error: Workspace {0} does not exist!".format(wrkspc)
        #    logging.error(msg)
        #    print msg
        #    sys.exit()

        # Read Excel Sheet
        wb = open_workbook(excel_doc)
        for sheet in wb.sheets():
            # Only use sheet 'AddFields'
            if sheet.name != 'DomainOperations':
              continue
            number_of_rows = sheet.nrows
            number_of_columns = sheet.ncols
            logging.info(' ')
            print ' '
            logging.info('Processing sheet: ' + sheet.name)
            print 'Processing sheet: ' + sheet.name
            if number_of_rows > 1:
              print "Excel Rows: " + str(number_of_rows)
              arcpy.env.workspace = wrkspc

        
              for row in range(1, number_of_rows):
                in_table = (sheet.cell(row,0).value)
                field_name = (sheet.cell(row,1).value)
                domain_name = (sheet.cell(row,2).value)
                domain_description = (sheet.cell(row,3).value)
                domain_field_type = (sheet.cell(row,4).value)
                subtype_code = (sheet.cell(row,5).value)
                operation = (sheet.cell(row,6).value)
                new_domain_name = (sheet.cell(row,7).value)
                coded_value_add = (sheet.cell(row,8).value)
                coded_desc_add = (sheet.cell(row,9).value)
                coded_value_remove = (sheet.cell(row,10).value)
                update_records = (sheet.cell(row,11).value)

                if operation == "Assign":
                    AssignDomain(in_table, field_name, domain_name, subtype_code)
                    continue

                if operation == "Add":
                    print "Add domain..."
                    AddDomain(wrkspc, domain_name, domain_description, domain_field_type)
                    continue

                if operation == "Add Code":
                    print "Add coded value..."
                    AddCodedValue(wrkspc, domain_name, coded_value_add, coded_desc_add)
                    continue

                if operation == "Remove":
                    print "Remove domain..."

                if operation == "Update Code":
                    UpdateDomainCode(wrkspc, in_table, field_name, domain_name, coded_value_add, coded_desc_add, coded_value_remove, update_records)
                
            else:
              logging.warning('No rows found in excel doc!')

        logging.info('Finished')
        print 'Finished'
 
    except Exception as err:
        import traceback, sys
        tb = sys.exc_info()[2]
        logging.error("Line %i" % tb.tb_lineno)
        print "Line %i" % tb.tb_lineno
        logging.error(err.message)
        print err.message
    


if __name__ == '__main__':
    main(sys.argv) 
    